import asyncio
import logging
from typing import Optional
import aiohttp

MAX_CONCURRENT_REQUESTS = 5
REQUEST_TIMEOUT = 10

class MBTASessionManager:
    """Singleton class to manage a shared aiohttp.ClientSession."""

    _session: Optional[aiohttp.ClientSession] = None
    _semaphore: Optional[asyncio.Semaphore] = None
    _logger: Optional[logging.Logger] = None
    _max_concurrent_requests: int = MAX_CONCURRENT_REQUESTS
    _timeout: int = REQUEST_TIMEOUT
    _own_session: bool = True

    @classmethod
    def configure(
        cls,
        session: Optional[aiohttp.ClientSession] = None,
        logger: Optional[logging.Logger] = None,
        max_concurrent_requests: Optional[int] = None,
        timeout: Optional[int] = None,
    ):
        """Configure the SessionManager."""
        cls._logger = logger or logging.getLogger(__name__)
        cls._max_concurrent_requests = max_concurrent_requests or cls._max_concurrent_requests
        cls._timeout = timeout or cls._timeout
        cls.semaphore = asyncio.Semaphore(cls._max_concurrent_requests)
        if session:
            cls._session = session
            cls._own_session = False
        cls._logger.debug("MBTASessionManager initialized")

    @classmethod
    async def get_session(cls) -> aiohttp.ClientSession:
        """Get the shared aiohttp.ClientSession instance, creating it if necessary."""
        if cls._session is None or cls._session.closed:
            cls._logger.debug("Creating a new aiohttp.ClientSession instance")
            cls._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=cls._timeout))
            cls._own_session = True
        return cls._session

    @classmethod
    async def close_session(cls):
        """Close the shared aiohttp.ClientSession."""
        if cls._own_session and cls._session and not cls._session.closed:
            try:
                cls._logger.debug("Closing the aiohttp.ClientSession instance")
                await cls._session.close()
            except Exception as e:
                if cls._logger:
                    cls._logger.error(f"Error closing session: {e}")
            finally:
                cls._session = None

    @classmethod
    async def cleanup(cls):
        """Clean up resources when shutting down."""
        cls._logger.debug("Cleaning up MBTASessionManager resources")
        await cls.close_session()
        cls.semaphore = None

    @classmethod
    async def __aenter__(cls):
        await cls.get_session()
        return cls

    @classmethod
    async def __aexit__(cls, exc_type, exc, tb):
        await cls.cleanup()



from dataclasses import dataclass
from enum import Enum
from typing import Optional
from datetime import datetime, timedelta

from .mbta_object_store import MBTAStopObjStore
from .models.mbta_stop import MBTAStop

class StopType(Enum):
    DEPARTURE = "departure"
    ARRIVAL = "arrival"

@dataclass
class Time:
    """
    Represents a time with optional original and updated values.
    """
    scheduled_time: Optional[datetime] = None
    predicted_time: Optional[datetime] = None

    def __init__(self, scheduled_time: Optional[datetime] = None):
        self.scheduled_time = scheduled_time
        self.predicted_time: Optional[datetime] = None

    @property
    def deltatime(self) -> Optional[timedelta]:
        if self.scheduled_time and self.predicted_time:
            return self.predicted_time - self.scheduled_time
        if self.scheduled_time:
            return timedelta(seconds=0)
        return None

    @property
    def time(self) -> Optional[datetime]:
        return self.predicted_time or self.scheduled_time

@dataclass
class Stop:
    """
    Represents a stop on a trip.
    """
    stop_type: StopType
    mbta_stop_id: str
    stop_sequence: int
    arrival: Optional[Time] = None
    departure: Optional[Time] = None
    status: Optional[str] = None

    def __init__(
        self,
        stop_type: StopType,
        mbta_stop_id: str,
        stop_sequence: int,
        arrival_time: Optional[datetime] = None,
        departure_time: Optional[datetime] = None,
        status: Optional[str] = None):
        """
        Inits the stop.

        Args:
            stop_type: the stop type (DEPARTURE|ARRIVAL)
            mbta_stop_id: The MBTA stop ID.
            stop_sequence: The stop sequence.
            arrival_time: The arrival time.
            departure_time: The departure time.
        """
        self.stop_type = stop_type
        self.mbta_stop_id = mbta_stop_id
        self.stop_sequence = stop_sequence
        self.arrival = Time(scheduled_time=arrival_time) if arrival_time else None
        self.departure = Time(scheduled_time=departure_time) if departure_time else None
        self.status = status

    @property
    def mbta_stop(self) -> Optional[MBTAStop]:
        """Retrieve the MBTAStop object for this TripStop."""
        mbta_stop = MBTAStopObjStore.get_by_id(self.mbta_stop_id)
        if mbta_stop:
            return mbta_stop
        return None

    @mbta_stop.setter
    def mbta_stop(self, mbta_stop: "MBTAStop") -> None:
        """Set the MBTAStop and add it to the registry."""
        self.mbta_stop_id = mbta_stop.id  # Update the stop ID
        MBTAStopObjStore.store(mbta_stop)  # Add to store

    @property
    def arrival_time(self) -> Optional[datetime]:
        if self.arrival and self.arrival.time:
            return self.arrival.time
        return None

    @property
    def departure_time(self) -> Optional[datetime]:
        if self.departure and self.departure.time:
            return self.departure.time
        return None

    @property
    def time(self) -> Optional[datetime]:
        """Returns the most recent time for this stop (updated or original)."""
        if self.arrival:
            return self.arrival.time
        elif self.departure:
            return self.departure.time
        else:
            return None

    @property
    def deltatime(self) -> Optional[timedelta]:
        if self.arrival and self.arrival.deltatime:
            return self.arrival.deltatime
        if self.departure and self.departure.deltatime:
            return self.departure.deltatime
        return None

    @property
    def time_to(self) -> Optional[timedelta]:
        if self.time:
            return self.time.astimezone() - datetime.now().astimezone()
        return None

    def __repr__(self) -> str:
        return (f"TripStop({self.stop_type.value}): {self.mbta_stop_id} @ {self.time.replace(tzinfo=None)}"
        )

    def update_stop(
        self,
        mbta_stop_id: str,
        stop_sequence: int,
        arrival_time: Optional[datetime] = None,
        departure_time: Optional[datetime] = None,
        status: Optional[str] = None) -> None:
        """
        Updates the stop with new information.

        Args:
            mbta_stop_id: The new MBTA stop ID.
            stop_sequence: The new stop sequence.
            arrival_time: The new arrival time.
            departure_time: The new departure time.
        """
        self.mbta_stop_id = mbta_stop_id
        self.stop_sequence = stop_sequence
        self.status = status

        if arrival_time:
            if not self.arrival:
                self.arrival = Time(scheduled_time=arrival_time)
            else:
                self.arrival.predicted_time = arrival_time
        if departure_time:
            if not self.departure:
                self.departure = Time(scheduled_time=departure_time)
            else:
                self.departure.predicted_time = departure_time



import asyncio
from datetime import datetime
from typing import Optional
import logging

from ..stop import StopType

from ..client.mbta_client import MBTAClient
from ..handlers.base_handler import MBTABaseHandler

from ..trip import Trip

class TripsHandler(MBTABaseHandler):
    """Handler for managing Trips."""

    DEFAULT_MAX_TRIPS =  5
    
    @classmethod
    async def create(
        cls, 
        departure_stop_name: str ,
        mbta_client: MBTAClient,
        arrival_stop_name: str,
        max_trips: Optional[int] = DEFAULT_MAX_TRIPS,
        sort_by: Optional[StopType] = StopType.ARRIVAL,
        logger: Optional[logging.Logger] = None)-> "TripsHandler":

        """Asynchronous factory method to initialize TripsHandler."""
        instance = await super()._create(
            departure_stop_name=departure_stop_name,
            mbta_client=mbta_client,
            arrival_stop_name=arrival_stop_name, 
            max_trips=max_trips,
            logger=logger)

        instance._sort_by = sort_by
        instance._logger = logger or logging.getLogger(__name__)  # Logger instance

        return instance

    async def update(self) -> list[Trip]:
        self._logger.debug("Updating trips scheduling and info")

        try:
            # Initialize trips
            trips: dict[str, Trip] = {}

            # round to the hour to leverage caching in fetch and processing (based on header hash)
            min_time = f"{datetime.now().strftime("%H")}:00"

            params = {
                'filter[min_time]': min_time
            }
            
            # Update trip scheduling
            updated_trips = await super()._update_scheduling(trips=trips,params=params)

            # Filter out departed trips
            filtered_trips = super()._filter_and_sort_trips(
                trips=updated_trips,
                remove_departed=True,
                sort_by=self._sort_by)
            
            # Update stops for the trip
            task_stops = asyncio.create_task(super()._update_mbta_stops_for_trips(trips=filtered_trips.values()))
            # Update trip details
            tasks_trips_details = asyncio.create_task(super()._update_details(trips=filtered_trips))
    
            await task_stops                             
            detailed_trips = await tasks_trips_details

            # Filter out departed trips again
            filtered_detailed_trips = super()._filter_and_sort_trips(
                trips=detailed_trips,
                remove_departed=True,
                sort_by=self._sort_by)

            # Limit trips to the maximum allowed
            limited_trips = dict(list(filtered_detailed_trips.items())[:self._max_trips])

            # Return the sorted trips as a list
            return list(limited_trips.values())

        except Exception as e:
            self._logger.error(f"Failed to update trips: {e}")
            return []



from datetime import datetime
import logging
from typing import Any, Optional

class MBTAVehicle:
    """A vehicle object to hold information about an MBTA vehicle."""

    def __init__(self, vehicle: dict[str, Any]) -> None:
        try:
            # ID
            self.id: str = vehicle.get('id', '')
            
            # Attributes
            attributes = vehicle.get('attributes', {})
            self.current_status: Optional[str] = attributes.get('current_status', None)
            self.current_stop_sequence: Optional[str] = attributes.get('current_stop_sequence', None)
            self.direction_id: Optional[str] = attributes.get('direction_id', None)
            self.label: Optional[str] = attributes.get('label', None) 
            self.occupancy_status: Optional[str] = attributes.get('occupancy_status', None)
            self.revenue: Optional[int] = attributes.get('revenue', None)
            self.speed: Optional[str] = attributes.get('speed', None)
            self.updated_at: Optional[datetime] = (
                datetime.fromisoformat(attributes.get('updated_at'))
                if attributes.get('updated_at') is not None
                else None
            )
            self.latitude: Optional[str] = attributes.get('latitude', None)
            self.longitude: Optional[str] = attributes.get('longitude', None)

            # Relationships
            relationships = vehicle.get('relationships', {})
            self.trip_id: Optional[str] = relationships.get('trip', {}).get('data', {}).get('id', None) if relationships.get('trip').get('data') else None
            self.stop_id: Optional[str] = relationships.get('stop', {}).get('data', {}).get('id', None) if relationships.get('stop').get('data') else None
            self.route_id: Optional[str] = relationships.get('route', {}).get('data', {}).get('id', None) if relationships.get('route').get('data') else None
            
        except Exception as e:
            # Log the exception with traceback
            logger = logging.getLogger(__name__)
            logger.error(f"Error initializing {self.__class__.__name__}: {e}", exc_info=True)

    def __repr__(self) -> str:
        return (f"MBTAVehicles:{self.id}")

    def __eq__(self, other: object) -> bool:
        """Defines equality based on the vehicle ID."""
        if isinstance(other, MBTAVehicle):
            return self.id == other.id
        return False




import asyncio
import logging
from datetime import datetime
from typing import Optional

from .base_handler import MBTABaseHandler
from ..client.mbta_client import MBTAClient
from ..trip import Trip


class DeparturesHandler(MBTABaseHandler):
    """Handler for managing timetable."""

    DEFAULT_MAX_TRIPS =  5

    @classmethod
    async def create(
        cls,
        mbta_client: MBTAClient,
        departure_stop_name: str ,
        max_trips: Optional[int] = DEFAULT_MAX_TRIPS,
        logger: Optional[logging.Logger] = None)-> "DeparturesHandler":

        """Asynchronous factory method to initialize DeparturesHandler."""

        instance = await super()._create(
            mbta_client=mbta_client,
            departure_stop_name=departure_stop_name,
            max_trips=max_trips,
            logger=logger)

        instance._logger = logger or logging.getLogger(__name__)  # Logger instance

        return instance

    async def update(self) -> list[Trip]:
        self._logger.debug("Updating Trips")
        try:

            # Initialize trips
            trips: dict[str, Trip] = {}

            # roudn to the hour to leverage caching in fetch and processing (based on header hash)
            min_time = f"{datetime.now().strftime("%H")}:00"

            params = {
                'filter[min_time]': min_time
            }

            # Update trip scheduling
            updated_trips = await super()._update_scheduling(trips=trips, params=params)

            # Filter out departed trips'
            filtered_trips = super()._filter_and_sort_trips(
                trips=updated_trips,
                remove_departed=True,
                require_both_stops=False)

            # Update stops for the trip
            task_stops = asyncio.create_task(
                super()._update_mbta_stops_for_trips(trips=filtered_trips.values()))
            # Update trip details
            tasks_trips_details = asyncio.create_task(
                super()._update_details(trips=filtered_trips))

            await task_stops
            detailed_trips = await tasks_trips_details

            # Filter out departed trips again
            filtered_trips = super()._filter_and_sort_trips(
                trips=detailed_trips,
                remove_departed=True,
                require_both_stops=False)

            # Limit trips to the maximum allowed
            limited_trips = dict(list(filtered_trips.items())[:self._max_trips])

            return list(limited_trips.values())

        except Exception as e:
            self._logger.error(f"Error updating trips: {e}")
            raise



from collections import OrderedDict
from threading import RLock
from typing import Generic, TypeVar, Optional

from .models.mbta_alert import MBTAAlert
from .models.mbta_route import MBTARoute
from .models.mbta_stop import MBTAStop
from .models.mbta_trip import MBTATrip
from .models.mbta_vehicle import MBTAVehicle

T = TypeVar("T")  # Generic type for objects stored in the object store


class MBTABaseObjStore(Generic[T]):
    """Base class for object stores without size limits."""
    _lock: RLock = RLock()  # Thread-safe lock

    @classmethod
    def get_by_id(cls, id: str) -> Optional[T]:
        """Retrieve an object by its ID and mark it as recently used."""
        with cls._lock:
            obj = cls._registry.get(id)
            if obj is not None:
                # Mark the object as recently used
                cls._registry.move_to_end(id)
            return obj

    @classmethod
    def store(cls, obj: T) -> None:
        """Add an object to the registry."""
        with cls._lock:
            # Use the object's `id` attribute for storage
            obj_id = getattr(obj, 'id', None)
            if not obj_id:
                raise ValueError("Object must have an 'id' attribute.")

            # Move existing item to the end or add a new one
            cls._registry[obj_id] = obj
            cls._registry.move_to_end(obj_id)

    @classmethod
    def clear_store(cls) -> None:
        """Clear all objects from the registry."""
        with cls._lock:
            cls._registry.clear()

    @classmethod
    def __len__(cls) -> int:
        """Return the current number of items in the registry."""
        with cls._lock:
            return len(cls._registry)


class MBTASizedObjStore(MBTABaseObjStore[T]):
    """Subclass of MBTABaseObjStore with a limit on the number of objects."""
    _max_items: int = 512  # Default maximum size of the registry

    @classmethod
    def configure_max_items(cls, max_items: int) -> None:
        """Configure the maximum size of the registry."""
        with cls._lock:
            cls._max_items = max_items

    @classmethod
    def store(cls, obj: T) -> None:
        """Add an object to the registry, enforcing size limits."""
        with cls._lock:
            obj_id = getattr(obj, 'id', None)
            if not obj_id:
                raise ValueError("Object must have an 'id' attribute.")

            if obj_id in cls._registry:
                # Move existing item to the end
                cls._registry.move_to_end(obj_id)
            elif len(cls._registry) >= cls._max_items:
                # Evict the oldest item if size limit is reached
                cls._registry.popitem(last=False)

            # Add or update the item
            cls._registry[obj_id] = obj


# Object Stores (choose base class based on the need)
class MBTARouteObjStore(MBTABaseObjStore[MBTARoute]):
    """Uncapped registry for MBTA Route objects."""
    _registry: OrderedDict[str, MBTARoute] = OrderedDict()


class MBTAStopObjStore(MBTABaseObjStore[MBTAStop]):
    """Uncapped registry for MBTA Stop objects."""
    _registry: OrderedDict[str, MBTAStop] = OrderedDict()

    @classmethod
    def get_by_child_stop_id(cls, child_stop_id: str) -> Optional[MBTAStop]:
        """Retrieve a stop that contains the given child_stop_id and mark it as recently used."""
        with cls._lock:
            for stop_id, stop in cls._registry.items():
                if child_stop_id ==  stop_id or child_stop_id in stop.child_stops:
                    # Move the found stop to the end to mark it as recently used
                    cls._registry.move_to_end(stop_id)
                    return stop

        return None  # Return None if no stop contains the child_stop_id


class MBTATripObjStore(MBTASizedObjStore[MBTATrip]):
    """Capped registry for MBTA Trip objects."""
    _registry: OrderedDict[str, MBTATrip] = OrderedDict()


class MBTAVehicleObjStore(MBTASizedObjStore[MBTAVehicle]):
    """Capped registry for MBTA Vehicle objects."""
    _registry: OrderedDict[str, MBTAVehicle] = OrderedDict()


class MBTAAlertObjStore(MBTASizedObjStore[MBTAAlert]):
    """Capped registry for MBTA Alert objects."""
    _registry: OrderedDict[str, MBTAAlert] = OrderedDict()



from datetime import datetime
import logging
from typing import Any, Optional
from enum import Enum

class MBTAAlert:
    """An alert object to hold information about an MBTA alert."""

    def __init__(self, alert: dict[str, Any]) -> None:
        try:
            # ID
            self.id: Optional[str] = alert.get('id', None)

            # Attributes
            attributes = alert.get('attributes', {})
            self.url: Optional[str] = attributes.get('url', None)
            self.updated_at: Optional[datetime] = (
                datetime.fromisoformat(attributes.get('updated_at'))
                if attributes.get('updated_at') is not None
                else None
            )
            self.timeframe: Optional[str] = attributes.get('timeframe', None)
            self.short_header: Optional[str] = attributes.get('short_header', None)
            self.severity: Optional[str] = attributes.get('severity', None)
            self.service_effect: Optional[str] = attributes.get('service_effect', None)
            self.lifecycle: Optional[str] = attributes.get('lifecycle', None)

            # Informed Entities
            self.informed_entities: list[MBTAAlertsInformedEntity] = [
                MBTAAlertsInformedEntity(
                    trip_id=entity.get('trip'),
                    stop_id=entity.get('stop'),
                    route_type=entity.get('route_type'),
                    route_id=entity.get('route'),
                    facility_id=entity.get('facility'),
                    direction_id=entity.get('direction_id'),
                    activities=entity.get('activities')
                )
                for entity in attributes.get('informed_entity', [])
            ]
            
            self.image_alternative_text: Optional[str] = attributes.get('image_alternative_text', None)
            self.image: Optional[str] = attributes.get('image', None)
            self.header: Optional[str] = attributes.get('header', None)
            self.effect_name: Optional[str] = attributes.get('effect_name', None)
            self.effect: Optional[str] = attributes.get("effect", None)
            self.duration_certainty: Optional[str] = attributes.get('duration_certainty', None)
            self.description: Optional[str] = attributes.get('description', None)
            self.created_at: Optional[datetime] = (
                datetime.fromisoformat(attributes.get('created_at'))
                if attributes.get('created_at') is not None
                else None
            )
            self.cause: Optional[str] = attributes.get('cause', None)            
            self.banner: Optional[str] = attributes.get('banner', None)           
            # Active period
            self.active_period_start: Optional[datetime] = (
                datetime.fromisoformat(attributes.get('active_period', [{}])[0].get('start'))
                if attributes.get('active_period', [{}])[0].get('start') is not None
                else None
            )
            self.active_period_end: Optional[datetime] = (
                datetime.fromisoformat(attributes.get('active_period', [{}])[0].get('end'))
                if attributes.get('active_period', [{}])[0].get('end') is not None
                else None
            )
        
        except Exception as e:
            # Log the exception with traceback
            logger = logging.getLogger(__name__)
            logger.error(f"Error initializing {self.__class__.__name__}: {e}", exc_info=True)
        
    def __repr__(self) -> str:
        return (f"MBTAAlert:{self.id}")

    def __eq__(self, other: object) -> bool:
        """Defines equality based on the alert ID."""
        if isinstance(other, MBTAAlert):
            return self.id == other.id
        return False
    
class MBTAAlertsInformedEntity:
    def __init__(
        self,
        trip_id: Optional[str] = None,
        stop_id: Optional[str] = None,
        route_type: Optional[int] = None,
        route_id: Optional[str] = None,
        facility_id: Optional[str] = None,
        direction_id: Optional[int] = None,
        activities: Optional[list[str]] = None
    ):
        self.trip_id = trip_id
        self.stop_id = stop_id
        self.route_type = route_type
        self.route_id = route_id
        self.facility_id = facility_id
        self.direction_id = direction_id
        self.activities = activities

from enum import Enum

class MBTAAlertPassengerActivity(Enum):
    BOARD = "BOARD"  # Boarding a vehicle. Any passenger trip includes boarding a vehicle and exiting from a vehicle.
    BRINGING_BIKE = "BRINGING_BIKE"  # Bringing a bicycle while boarding or exiting.
    EXIT = "EXIT"  # Exiting from a vehicle (disembarking). Any passenger trip includes boarding a vehicle and exiting a vehicle.
    PARK_CAR = "PARK_CAR"  # Parking a car at a garage or lot in a station.
    RIDE = "RIDE"  # Riding through a stop without boarding or exiting. Not every passenger trip will include this – a passenger may board at one stop and exit at the next stop.
    STORE_BIKE = "STORE_BIKE"  # Storing a bicycle at a station.
    USING_ESCALATOR = "USING_ESCALATOR"  # Using an escalator while boarding or exiting (should only be used for customers who specifically want to avoid stairs.)
    USING_WHEELCHAIR = "USING_WHEELCHAIR"  # Using a wheelchair while boarding or exiting. Note that this applies to something that specifically affects customers who use a wheelchair to board or exit; a delay should not include this as an affected activity unless it specifically affects customers using wheelchairs.



import logging
from typing import Any, Optional

class MBTAStop:
    """A stop object to hold information about a stop."""

    def __init__(self, stop: dict[str, Any]) -> None:
        try:
            # ID
            self.id: str = stop.get('id', '')
            
            # Attributes
            attributes = stop.get('attributes', {})
            self.address: Optional[str] = attributes.get('address', None)
            self.at_street: Optional[str] = attributes.get('at_street', None)
            self.description: Optional[str] = attributes.get('description', None)
            self.latitude: Optional[float] = attributes.get('latitude', None)
            self.location_type: Optional[int] = attributes.get('location_type', None)
            self.longitude: Optional[float] = attributes.get('longitude', None)
            self.municipality: Optional[str] = attributes.get('municipality', None)
            self.name: Optional[str] = attributes.get('name', None)
            self.on_street: Optional[str] = attributes.get('on_street', None)
            self.platform_code: Optional[str] = attributes.get('platform_code', None)
            self.platform_name: Optional[str] = attributes.get('platform_name', None)
            self.vehicle_type: Optional[int] = attributes.get('vehicle_type', None)
            self.wheelchair_boarding: Optional[int] = attributes.get('wheelchair_boarding', None)

            # Extract child stops
            relationships = stop.get("relationships", {})
            child_stops_data = relationships.get("child_stops", {}).get("data", []) if relationships.get('child_stops', {}).get('data') else None
            self.child_stops: list[str] = [child["id"] for child in child_stops_data] if child_stops_data else []

        except Exception as e:
            # Log the exception with traceback
            logger = logging.getLogger(__name__)
            logger.error(f"Error initializing {self.__class__.__name__}: {e}", exc_info=True)
        
    def __repr__(self) -> str:
        return (f"MBTAStop:{self.id}")

    def __eq__(self, other: object) -> bool:
        """Defines equality based on the stop ID."""
        if isinstance(other, MBTAStop):
            return self.id == other.id
        return False



from .models.mbta_alert import MBTAAlert
from .models.mbta_prediction import MBTAPrediction
from .models.mbta_route import MBTARoute
from .models.mbta_schedule import MBTASchedule
from .models.mbta_stop import MBTAStop
from .models.mbta_trip import MBTATrip
from .models.mbta_vehicle import MBTAVehicle
from .client.mbta_client import MBTAClient
from .trip import Trip
from .stop import Stop, StopType
from .handlers.trips_handler import TripsHandler
from .handlers.trains_handler import TrainsHandler
from .handlers.departures_handler import DeparturesHandler

__all__ = [
    "MBTAAlert",
    "MBTAPrediction",
    "MBTARoute",
    "MBTASchedule",
    "MBTAStop",
    "MBTATrip",
    "MBTAVehicle",
    "MBTAClient",
    "Trip",
    "Stop",
    "StopType",
    "TripsHandler",
    "TrainsHandler",
    "DeparturesHandler"
]


from dataclasses import dataclass, field
from typing import Union, Optional
from datetime import datetime

from .mbta_object_store import MBTAAlertObjStore, MBTARouteObjStore, MBTAStopObjStore, MBTATripObjStore, MBTAVehicleObjStore

from .stop import Stop, StopType

from .models.mbta_schedule import MBTASchedule
from .models.mbta_prediction import MBTAPrediction
from .models.mbta_route import MBTARoute
from .models.mbta_trip import MBTATrip
from .models.mbta_vehicle import MBTAVehicle
from .models.mbta_alert import MBTAAlert


@dataclass
class Trip:
    """A class to manage a Trip with multiple stops."""
    _mbta_route_id: Optional[str] = None
    _mbta_trip_id: Optional[str] = None
    _mbta_vehicle_id: Optional[str] = None
    _mbta_alerts_ids: set[Optional[str]] = field(default_factory=set)
    stops: list[Optional['Stop']] = field(default_factory=list)

    VEHICLE_DATA_FRESHNESS_THRESHOLD = 90  # seconds

    # registry
    @property
    def mbta_route(self) -> Optional[MBTARoute]:
        """Retrieve the MBTARoute object for this Trip."""
        mbta_route = MBTARouteObjStore.get_by_id(self._mbta_route_id)
        if mbta_route:
            return mbta_route
        #self._mbta_route_id = None
        return None

    @mbta_route.setter
    def mbta_route(self, mbta_route: MBTARoute) -> None:
        if mbta_route:
            self._mbta_route_id = mbta_route.id
            MBTARouteObjStore.store(mbta_route)

    @property
    def mbta_trip(self) -> Optional[MBTATrip]:
        """Retrieve the MBTARoute object for this Trip."""
        mbta_trip = MBTATripObjStore.get_by_id(self._mbta_trip_id)
        if mbta_trip:
            return mbta_trip
        return None

    @mbta_trip.setter
    def mbta_trip(self, mbta_trip: MBTATrip) -> None:
        if mbta_trip:
            self._mbta_trip_id = mbta_trip.id
            MBTATripObjStore.store(mbta_trip)

    @property
    def mbta_vehicle(self) -> Optional[MBTAVehicle]:
        """Retrieve the MBTARoute object for this Trip."""
        return MBTAVehicleObjStore.get_by_id(self._mbta_vehicle_id)

    @mbta_vehicle.setter
    def mbta_vehicle(self, mbta_vehicle: MBTAVehicle) -> None:
        if mbta_vehicle:
            self._mbta_vehicle_id = mbta_vehicle.id
            MBTAVehicleObjStore.store(mbta_vehicle)

    @property
    def mbta_alerts(self) -> Optional[list[MBTAAlert]]:
        """Retrieve the MBTARoute object for this Trip."""
        mbta_alerts = []
        for mbta_alert_id in self._mbta_alerts_ids:
            mbta_alerts.append(MBTAAlertObjStore.get_by_id(mbta_alert_id))
        return mbta_alerts

    @mbta_alerts.setter
    def mbta_alerts(self, mbta_alerts: list[MBTAAlert]) -> None:
        if mbta_alerts:
            for mbta_alert in mbta_alerts:
                self._mbta_alerts_ids.add(mbta_alert.id)
                MBTAAlertObjStore.store(mbta_alert)

    # trip
    @property
    def headsign(self) -> Optional[str]:
        return self.mbta_trip.headsign if self.mbta_trip and self.mbta_trip.headsign else None

    @property
    def name(self) -> Optional[str]:
        return self.mbta_trip.name if self.mbta_trip and self.mbta_trip.name else None

    @property
    def destination(self) -> Optional[str]:
        return (
            self.mbta_route.direction_destinations[self.mbta_trip.direction_id]
            if self.mbta_trip
            and self.mbta_trip.direction_id
            and self.mbta_route
            and self.mbta_route.direction_destinations
            else None
        )

    @property
    def direction(self) -> Optional[str]:
        return (
            self.mbta_route.direction_names[self.mbta_trip.direction_id]
            if self.mbta_trip
            and self.mbta_trip.direction_id
            and self.mbta_route
            and self.mbta_route.direction_names
            else None
        )

    @property
    def duration(self) -> Optional[int]:
        if self._departure_stop and self._arrival_stop:
            return int((self._arrival_stop.time -  self._departure_stop.time).total_seconds())
        return None

    # route
    @property
    def route_name(self) -> Optional[str]:
        if self.mbta_route and self.mbta_route.type in [0,1,2,4]: #subway + train + ferry
            return self.mbta_route.long_name if self.mbta_route and self.mbta_route.long_name else None
        elif self.mbta_route and self.mbta_route.type == 3: #bus
            return self.mbta_route.short_name if self.mbta_route and self.mbta_route.short_name else None

    @property
    def route_color(self) -> Optional[str]:
        return f"#{self.mbta_route.color}" if self.mbta_route and self.mbta_route.color else None

    @property
    def route_description(self) -> Optional[str]:
        return MBTARoute.get_route_type_desc_by_type_id(self.mbta_route.type) if self.mbta_route and self.mbta_route.type is not None else None

    # vehicle
    @property
    def vehicle_status(self) -> Optional[str]:
        if self.mbta_vehicle and self.mbta_vehicle.current_status and self.vehicle_stop_name:
            title_case_with_spaces = " ".join([word.capitalize() for word in self.mbta_vehicle.current_status.split("_")])
            return title_case_with_spaces + " " + self.vehicle_stop_name
        return None

    @property
    def vehicle_stop_name(self) -> Optional[str]:
        return MBTAStopObjStore.get_by_child_stop_id(self.mbta_vehicle.stop_id).name if self.mbta_vehicle and self.mbta_vehicle.stop_id and MBTAStopObjStore.get_by_child_stop_id(self.mbta_vehicle.stop_id) else None

    @property
    def vehicle_longitude(self) -> Optional[float]:
        return self.mbta_vehicle.longitude if self.mbta_vehicle and self.mbta_vehicle.longitude else None

    @property
    def vehicle_latitude(self) -> Optional[float]:
        return self.mbta_vehicle.latitude if self.mbta_vehicle and self.mbta_vehicle.latitude else None

    @property
    def vehicle_occupancy(self) -> Optional[str]:
        return self.mbta_vehicle.occupancy_status if self.mbta_vehicle and self.mbta_vehicle.occupancy_status else None

    @property
    def vehicle_speed(self) -> Optional[str]:
        return self.mbta_vehicle.speed if self.mbta_vehicle and self.mbta_vehicle.speed else None

    @property
    def vehicle_updated_at(self) -> Optional[datetime]:
        return self.mbta_vehicle.updated_at.replace(tzinfo=None) if self.mbta_vehicle and self.mbta_vehicle.updated_at else None

    @property
    def is_vehicle_data_fresh(self) -> bool:
        if self.mbta_vehicle and self.mbta_vehicle.updated_at:
            now =  datetime.now().astimezone() # Ensure consistent timezone handling
            delta = (now - self.mbta_vehicle.updated_at).total_seconds()
            return delta <= self.VEHICLE_DATA_FRESHNESS_THRESHOLD
        return False

    #departure stop
    @property
    def _departure_stop(self) -> Optional[Stop]:
        return self.get_stop_by_type(StopType.DEPARTURE) if self.get_stop_by_type(StopType.DEPARTURE) else None

    @property
    def departure_stop_name(self) -> Optional[str]:
        return self._departure_stop.mbta_stop.name if self._departure_stop and self._departure_stop.mbta_stop else None

    @property
    def departure_platform(self) -> Optional[str]:
        return self._departure_stop.mbta_stop.platform_name if self._departure_stop and self._departure_stop.mbta_stop else None

    @property
    def departure_time(self) -> Optional[datetime]:
        return self._departure_stop.time.replace(tzinfo=None) if self._departure_stop and self._departure_stop.time else None

    @property
    def departure_delay(self) -> Optional[int]:
        return int(self._departure_stop.deltatime.total_seconds()) if self._departure_stop and self._departure_stop.deltatime else None

    @property
    def departure_time_to(self) -> Optional[int]:
        return int(self._departure_stop.time_to.total_seconds()) if self._departure_stop and self._departure_stop.time_to else None

    @property
    def departure_mbta_countdown(self) -> Optional[str]:
        return self._get_stop_mbta_countdown(StopType.DEPARTURE) if self._departure_stop else None

    @property
    def departure_countdown(self) -> Optional[str]:
        return self._get_stop_countdown(StopType.DEPARTURE) if self._departure_stop else None

    @property
    def has_departed(self) -> bool:
        stop = self._departure_stop
        if not stop:  # No departure stop available
            return False

        now = datetime.now().astimezone()
        seconds = (stop.departure_time.astimezone() - now).total_seconds() if stop.departure_time else stop.time
        current_stop = self.mbta_vehicle.current_stop_sequence if self.mbta_vehicle else None

        if current_stop is not None and current_stop > stop.stop_sequence:
            return True
        if seconds < -30:
            return True

        return False


    #arrival stop
    @property
    def _arrival_stop(self) -> Optional[Stop]:
        return self.get_stop_by_type(StopType.ARRIVAL) if self.get_stop_by_type(StopType.ARRIVAL) else None

    @property
    def arrival_stop_name(self) -> Optional[str]:
        return self._arrival_stop.mbta_stop.name if self._arrival_stop and self._arrival_stop.mbta_stop else None

    @property
    def arrival_platform(self) -> Optional[str]:
        return self._arrival_stop.mbta_stop.platform_name if self._arrival_stop and self._arrival_stop.mbta_stop else None

    @property
    def arrival_time(self) -> Optional[datetime]:
        return self._arrival_stop.time.replace(tzinfo=None) if self._arrival_stop and self._arrival_stop.time else None

    @property
    def arrival_delay(self) -> Optional[int]:
        return int(self._arrival_stop.deltatime.total_seconds()) if self._arrival_stop and self._arrival_stop.deltatime else None

    @property
    def arrival_time_to(self) -> Optional[int]:
        return int(self._arrival_stop.time_to.total_seconds()) if self._arrival_stop and self._arrival_stop.time_to else None

    @property
    def arrival_mbta_countdown(self) -> Optional[str]:
        return self._get_stop_mbta_countdown(StopType.ARRIVAL) if self._arrival_stop else None

    @property
    def arrival_countdown(self) -> Optional[str]:
        return self._get_stop_countdown(StopType.ARRIVAL) if self._arrival_stop else None

    @property
    def has_arrived(self) -> bool:
        stop = self._arrival_stop
        if not stop:  # No departure stop available
            return False

        now = datetime.now().astimezone()
        seconds = (stop.arrival_time.astimezone() - now).total_seconds() if stop.arrival_time else stop.time

        current_stop = self.mbta_vehicle.current_stop_sequence if self.mbta_vehicle else None

        if current_stop is not None and current_stop > stop.stop_sequence:
            return True
        if seconds < -30:
            return True

        return False

    #alerts
    @property
    def alerts(self) -> Optional[set[str]]:
        alerts_details = set()
        if self.mbta_alerts:
            for mbta_alert in self.mbta_alerts:
                effect = " ".join(mbta_alert.effect.split("_"))
                lifecycle = mbta_alert.lifecycle
                short_header = mbta_alert.short_header
                alerts_details.add(effect + "[" + lifecycle + "]: " + short_header)
            return alerts_details
        return None

    def get_stop_by_type(self, stop_type: str) -> Optional[Stop]:
        return next((stop for stop in self.stops if stop and stop.stop_type == stop_type), None)

    def add_stop(self, stop_type: str, scheduling: Union[MBTASchedule, MBTAPrediction], mbta_stop_id: str) -> None:
        """Add or update a stop in the journey."""
        stop = self.get_stop_by_type(stop_type)

        ##Status from prediction
        status = None
        if isinstance(scheduling, MBTAPrediction) and scheduling.status:
            status = scheduling.status

        if stop is None:
            # Create a new Stop
            stop = Stop(
                stop_type=stop_type,
                mbta_stop_id=mbta_stop_id,
                stop_sequence=scheduling.stop_sequence,
                arrival_time=scheduling.arrival_time,
                departure_time=scheduling.departure_time,
                status=status
            )
            self.stops.append(stop)
        else:
            # Update existing Stop
            stop.update_stop(
                mbta_stop_id=mbta_stop_id,
                stop_sequence=scheduling.stop_sequence,
                arrival_time=scheduling.arrival_time,
                departure_time=scheduling.departure_time,
                status=status
            )

    def remove_stop_by_id(self, mbta_stop_id: str) -> None:
        self.stops = [stop for stop in self.stops if stop.mbta_stop.id != mbta_stop_id]

    def reset_stops(self):
        self.stops = []

    def get_stop_id_by_stop_type(self, stop_type: StopType) -> Optional[str]:
        """Return the stop ID of the stop of the given type."""
        if stop_type == StopType.DEPARTURE and self._departure_stop and self._departure_stop.mbta_stop:
            return self._departure_stop.mbta_stop.id
        if stop_type == StopType.ARRIVAL and self._arrival_stop and self._arrival_stop.mbta_stop:
            return self._arrival_stop.mbta_stop.id
        return None

    def get_stops_ids(self) -> list[str]:
        """Return IDs of departure and arrival stops, excluding None."""
        return [
            stop_id for stop_id in [
                self.get_stop_id_by_stop_type(StopType.DEPARTURE),
                self.get_stop_id_by_stop_type(StopType.ARRIVAL)
            ] if stop_id is not None
        ]

    def get_alert_header(self, alert_index: int) -> Optional[str]:
        if 0 <= alert_index < len(self.mbta_alerts):
            return self.mbta_alerts[alert_index].header
        return None

    def _get_stop_countdown(self, stop_type: StopType) -> Optional[str]:
        """Determine the countdown or status of a stop."""

        stop: Stop = self.get_stop_by_type(stop_type)
        if not stop:
            return None

        if stop.status:
            return stop.status

        if not stop.time:
            return None

        now = datetime.now().astimezone()

        # Calculate time deltas
        arrival_time = stop.arrival_time or stop.time
        departure_time = stop.departure_time or stop.time

        arrival_delta = arrival_time.astimezone() - now
        departure_delta = departure_time.astimezone() - now

        seconds_arrival = int(arrival_delta.total_seconds())
        seconds_departure = int(departure_delta.total_seconds())

        # Vehicle data overrides
        if self.mbta_vehicle and self.is_vehicle_data_fresh:

            vehicle_stop = self.mbta_vehicle.current_stop_sequence
            vehicle_status = self.mbta_vehicle.current_status

            if vehicle_stop > stop.stop_sequence:
                if stop_type == StopType.DEPARTURE:
                    return "Departed"
                if stop_type == StopType.ARRIVAL:
                    return "Arrived"

            if vehicle_stop == stop.stop_sequence:
                if stop_type == StopType.DEPARTURE and vehicle_status == "STOPPED_AT" and 0 <= seconds_departure <= 90:
                    return "Boarding"
                if vehicle_status == "INCOMING_AT" and 0 <= seconds_arrival <= 90:
                    return "Arriving"

        # Handle status messages
        if stop_type == StopType.DEPARTURE and seconds_departure < 0:
            return "Departed"
        if stop_type == StopType.ARRIVAL and seconds_arrival < 0:
            return "Arrived"

        # Prioritize boarding over arriving
        if stop_type == StopType.DEPARTURE and 0 <= seconds_departure < 30:
            return "Boarding"
        if 0 <= seconds_arrival < 30 and seconds_departure > 0:
            return "Arriving"

        # Default to formatted time countdown
        return self._format_time(seconds_arrival) if seconds_arrival >= 30 else None

    def _get_stop_mbta_countdown(self, stop_type: StopType) -> Optional[str]:
        """Determine the countdown to a stop based on vehicle and time following
        https://www.mbta.com/developers/v3-api/best-practices """

        stop = self.get_stop_by_type(stop_type)

        if stop:

            if stop.status:
                return stop.status

            if not stop.time:
                return None

            seconds = (stop.time.astimezone() - datetime.now().astimezone()).total_seconds()

            if seconds < 0:
                return None

            if seconds <= 90 and self.mbta_vehicle and self.mbta_vehicle.current_status == "STOPPED_AT" and self.mbta_vehicle.current_stop_sequence == stop.stop_sequence:
                return "BRD"

            if seconds <= 30:
                return "ARR"

            if seconds <= 60:
                return "1 min"

            minutes = int(seconds/60)

            if minutes > 20:
                return "20+ min"

            return f"{minutes} min"

        return None

   # Convert time to human-readable format
    def _format_time(self, seconds: int) -> Optional[str]:
        if seconds < 0:
            return None
        minutes = (seconds // 60) % 60
        hours = (seconds // 3600) % 24
        days = seconds // 86400

        if days > 0:
            return f"{days}d {hours}h {minutes}m"
        if hours > 0:
            return f"{hours}h {minutes}m"
        if minutes > 1:
            return f"{minutes} min"
        return "1 min"



import aiohttp
import logging
from typing import Optional, Any, Dict, Tuple

from ..models.mbta_vehicle import MBTAVehicle
from ..models.mbta_route import MBTARoute
from ..models.mbta_stop import MBTAStop
from ..models.mbta_schedule import MBTASchedule
from ..models.mbta_prediction import MBTAPrediction
from ..models.mbta_trip import MBTATrip
from ..models.mbta_alert import MBTAAlert
from .mbta_cache_manager import MBTACacheManager, CacheEvent
from .mbta_session_manager import MBTASessionManager

MBTA_DEFAULT_HOST = "api-v3.mbta.com"

ENDPOINTS = {
    'STOPS': 'stops',
    'ROUTES': 'routes',
    'PREDICTIONS': 'predictions',
    'SCHEDULES': 'schedules',
    'TRIPS': 'trips',
    'ALERTS': 'alerts',
    'VEHICLES': 'vehicles', 
}



class MBTAClient:
    """Class to interact with the MBTA v3 API."""
    
    def __init__(self, 
                 api_key: Optional[str] = None,
                 session: Optional[aiohttp.ClientSession] = None, 
                 cache_manager: Optional[MBTACacheManager] = None, 
                 logger: Optional[logging.Logger] = None, 
                 max_concurrent_requests: Optional[int] = None,
                 timeout: Optional[int] = None):
        self._api_key = api_key

        self._logger = logger or logging.getLogger(__name__)
        
        MBTASessionManager.configure(
            session=session,
            logger=logger,
            max_concurrent_requests=max_concurrent_requests,
            timeout=timeout,
        )
        
        if cache_manager is None:
            self._cache_manager = MBTACacheManager()
            self._own_cache = True
        else:
            self._cache_manager = cache_manager
            self._own_cache = False
                    
        self._logger.debug("MBTAClient initialized")

    def __repr__(self) -> str:
        return (f"MBTAClient(own_cache={self._own_cache})"
        )
        
    async def __aenter__(self):
        """Enter the context and prepare the session."""
        await MBTASessionManager.get_session()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """Exit the context and clean up."""
        if self._own_cache == True:
            self._cache_manager.cleanup()
        await MBTASessionManager.cleanup()

    async def _fetch_data(
        self, path: str, params: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """Helper method to fetch data from the MBTA API."""
        try:
            data, timestamp = await self.request("GET", path, params)
            if "data" not in data:
                self._logger.warning(f"Response missing 'data': {data}")
                raise MBTAClientError("Invalid response from API.")
            return data, timestamp
        
        except MBTAClientError as error:
            self._logger.error(
                "MBTAClientError: %s (HTTP %s - %s) | URL: %s",
                error.message, error.status_code, error.reason, error.url
            )
            raise

        except Exception as error:
            self._logger.error("Unexpected error: %s", error)
            raise

    async def request(self, method: str, path: str, params: Optional[Dict[str, Any]] = None) -> Tuple[Any,float]:
        """Make an HTTP request with optional query parameters."""
        session = await MBTASessionManager.get_session()
        params = params or {}
        headers = {}

        if self._api_key:
            params["api_key"] = self._api_key

        url = f"https://{MBTA_DEFAULT_HOST}/{path}"
        
        cached_data, timestamp, last_modified = self._cache_manager.get_cached_data(path, params)
        if last_modified:
            headers["If-Modified-Since"] = last_modified
        headers["Accept-Encoding"] = "gzip"

        try:
            async with MBTASessionManager.semaphore:
                
                self._logger.debug(f"{method} {url} {headers} {params}")                
                response: aiohttp.ClientResponse = await session.request(
                    method, url, 
                    params=params, 
                    headers=headers
                )

                if response.status == 403:
                    self._logger.error("Authentication error: Invalid API key (HTTP 403).")
                    raise MBTAAuthenticationError("Invalid API key or credentials (HTTP 403).")
                
                if response.status == 429:
                    self._logger.error("Rate limit exceeded (HTTP 429).")
                    raise MBTATooManyRequestsError("Rate limit exceeded (HTTP 429). Please uscheck your API key.")

                if response.status == 304:
                    if cached_data is not None:
                        if  self._cache_manager.stats:
                            self._cache_manager.cache_stats.increase_counter(CacheEvent.HIT)
                        return cached_data, timestamp
                    else:
                        raise MBTAClientError(f"Cache empty for 304 response: {url}")

                response.raise_for_status()
                
                data = await response.json()
                last_modified = response.headers.get("Last-Modified")
                if last_modified:
                    timestamp = self._cache_manager.update_cache(path=path, params=params, data=data, last_modified=last_modified)
                if  self._cache_manager.stats:
                    self._cache_manager.cache_stats.increase_counter(CacheEvent.MISS)
                return data, timestamp

        except TimeoutError as error:
            self._logger.error(f"Timeout during request to {url}: {error}")
            raise TimeoutError
        
        except Exception as error:
            self._logger.error(f"Error during request to {url}: {error}")
            raise MBTAClientError("Unexpected error during request.") from error

    async def fetch_route(self, id: str, params: Optional[Dict[str, Any]] = None) -> Tuple[MBTARoute, float]:
        """Fetch a MBTARoute by its ID."""
        self._logger.debug(f"Fetching MBTA route with ID: {id}")
        data, timestamp = await self._fetch_data(f"{ENDPOINTS['ROUTES']}/{id}", params)
        return MBTARoute(data["data"]), timestamp

    async def fetch_trip(self, id: str, params: Optional[Dict[str, Any]] = None) -> Tuple[MBTATrip, float]:
        """Fetch a MBTATrip by its ID."""
        self._logger.debug(f"Fetching MBTA trip with ID: {id}")
        data, timestamp = await self._fetch_data(f"{ENDPOINTS['TRIPS']}/{id}", params)
        return MBTATrip(data["data"]), timestamp
    
    async def fetch_stop(self, id: str, params: Optional[Dict[str, Any]] = None) -> Tuple[MBTAStop, float]:
        """Fetch a MBTAStop by its ID."""
        self._logger.debug(f"Fetching MBTA stop with ID: {id}")
        data, timestamp = await self._fetch_data(f'{ENDPOINTS["STOPS"]}/{id}', params)
        return MBTAStop(data['data']), timestamp
    
    async def fetch_vehicle(self, id: str, params: Optional[Dict[str, Any]] = None) -> Tuple[MBTAVehicle, float]:
        """Fetch a MBTAVehicle by its ID."""
        self._logger.debug("Fetching MBTA vehicle with ID: {id}")
        data, timestamp = await self._fetch_data(ENDPOINTS['VEHICLES']/{id}, params)
        return MBTAVehicle(data['data']), timestamp

    async def fetch_routes(self, params: Optional[Dict[str, Any]] = None) -> Tuple[list[MBTARoute], float]:
        """Fetch a list of MBTARoute."""
        self._logger.debug("Fetching all MBTA routes")
        data, timestamp = await self._fetch_data(ENDPOINTS["ROUTES"], params)
        return [MBTARoute(item) for item in data["data"]], timestamp

    async def fetch_trips(self, params: Optional[Dict[str, Any]] = None) -> Tuple[list[MBTATrip], float]:
        """Fetch a list of MBTATrip."""
        self._logger.debug("Fetching MBTA trips")
        data, timestamp = await self._fetch_data(ENDPOINTS["TRIPS"], params)
        return [MBTATrip(item) for item in data["data"]], timestamp

    async def fetch_stops(self, params: Optional[Dict[str, Any]] = None) -> Tuple[list[MBTAStop], float]:
        """Fetch a list of MBTAStops."""
        self._logger.debug("Fetching MBTA stops")
        data, timestamp = await self._fetch_data(ENDPOINTS['STOPS'], params)
        return [MBTAStop(item) for item in data["data"]], timestamp

    async def fetch_schedules(self, params: Optional[Dict[str, Any]] = None) -> Tuple[list[MBTASchedule], float]:
        """Fetch a list of MBTASchedules."""
        self._logger.debug("Fetching MBTA schedules")
        data, timestamp = await self._fetch_data(ENDPOINTS['SCHEDULES'], params)
        return [MBTASchedule(item) for item in data["data"]], timestamp

    async def fetch_vehicles(self, params: Optional[Dict[str, Any]] = None) -> Tuple[list[MBTAVehicle], float]:
        """Fetch a list of MBTAAlerts."""
        self._logger.debug("Fetching MBTA vehicles")
        data, timestamp = await self._fetch_data(ENDPOINTS['VEHICLES'], params)
        return [MBTAVehicle(item) for item in data["data"]], timestamp

    async def fetch_predictions(self, params: Optional[Dict[str, Any]] = None) -> Tuple[list[MBTAPrediction], float]:
        """Fetch a list of MBTAPredictions."""
        self._logger.debug("Fetching MBTA predictions")
        data, timestamp = await self._fetch_data(ENDPOINTS['PREDICTIONS'], params)
        return [MBTAPrediction(item) for item in data["data"]], timestamp

    async def fetch_alerts(self, params: Optional[Dict[str, Any]] = None) -> Tuple[list[MBTAAlert], float]:
        """Fetch a list of MBTAAlerts."""
        self._logger.debug("Fetching MBTA alerts")
        data, timestamp = await self._fetch_data(ENDPOINTS['ALERTS'], params)
        return [MBTAAlert(item) for item in data["data"]], timestamp

class MBTAAuthenticationError(Exception):
    """Custom exception for MBTA authentication errors."""

class MBTATooManyRequestsError(Exception):
    """Custom exception for MBTA TooManyRequests errors."""

class MBTAClientError(Exception):
    """Custom exception for MBTA client errors."""
    def __init__(self, message, status_code=None, reason=None, url=None):
        self.message = message
        self.status_code = status_code
        self.reason = reason
        self.url = url
        super().__init__(self.__str__())

    def __str__(self):
        base_message = f"MBTAClientError: {self.message}"
        if self.status_code:
            base_message += f" (HTTP {self.status_code} - {self.reason})"
        if self.url:
            base_message += f" | URL: {self.url}"
        return base_message



import asyncio
import logging
from datetime import datetime, timedelta
from typing import Optional, Tuple

from ..handlers.base_handler import MBTABaseHandler
from ..client.mbta_client import MBTAClient
from ..mbta_object_store import MBTATripObjStore
from ..models.mbta_trip import MBTATrip
from ..trip import Trip

class TrainsHandler(MBTABaseHandler):
    """Handler for managing Trips."""

    DEFAULT_MAX_TRIPS = 1
    
    @classmethod
    async def create(
        cls,
        mbta_client: MBTAClient,
        departure_stop_name: str ,
        arrival_stop_name: str,
        trips_name: str,
        max_trips: Optional[int] = DEFAULT_MAX_TRIPS,
        logger: Optional[logging.Logger] = None)-> "TrainsHandler":

        """Asynchronous factory method to initialize TripsHandler."""
        instance = await super()._create(
            mbta_client=mbta_client,
            departure_stop_name=departure_stop_name,
            arrival_stop_name=arrival_stop_name,
            max_trips=max_trips,
            logger=logger)
        
        instance._logger = logger or logging.getLogger(__name__)  # Logger instance

        await instance.__update_mbta_trips_by_trip_name(trips_name)

        return instance

    async def __update_mbta_trips_by_trip_name(self, trips_name: str) -> None:
        self._logger.debug("Updating MBTA trips")
        try:
            mbta_trips, _ = await self.__fetch_trips_by_name(trips_name)
            if mbta_trips:
                for mbta_trip in mbta_trips:
                    if not MBTATripObjStore.get_by_id(mbta_trip.id):
                        MBTATripObjStore.store(mbta_trip)
                    self._mbta_trips_id = mbta_trip.id
            else:
                self._logger.error(f"Invalid MBTA trip name {trips_name}")
                raise MBTATripError(f"Invalid MBTA trip name {trips_name}")
  
        except Exception as e:
            self._logger.error(f"Error updating MBTA trips: {e}")
            raise

    async def __fetch_trips_by_name(self, train_name: str) -> Tuple[list[MBTATrip],float]:    

        params = {
            'filter[revenue]': 'REVENUE',
            'filter[name]': train_name
            }

        mbta_trips, timestamp = await self._mbta_client.fetch_trips(params)
        return mbta_trips, timestamp


    async def update(self) -> list[Trip]:
        self._logger.debug("Updating trips scheduling and info")
        try:

            now = datetime.now().astimezone()

            # Initialize trips
            weekly_trips: list[dict[str, Trip]] = []

            for i in range(7):
                daily_trip: dict[str, Trip] = {}
                date_to_try = (now + timedelta(days=i)).strftime('%Y-%m-%d')

                params = {
                    'filter[trip]': self._mbta_trips_id,
                    'filter[date]': date_to_try
                }

                daily_updated_trip = await super()._update_scheduling(trips=daily_trip,params=params)

                # Filter out departed trips
                daily_filtered_trip = super()._filter_and_sort_trips(
                    trips=daily_updated_trip,
                    remove_departed=False)
                
                if len(daily_filtered_trip) > 0:
                    weekly_trips.append(daily_filtered_trip)
                
                if len(weekly_trips) == self._max_trips:
                    break
                
                if len(weekly_trips) == 0:
                    if i == 6:
                        self._logger.error(f"No trips between the provided stops till {date_to_try}")
                        raise MBTATripError(f"No trips between the provided stops till {date_to_try}")
                    continue

            trains: list[Trip] = []
            for trips in weekly_trips:
                
                # Update stops for the trip
                task_stops = asyncio.create_task(super()._update_mbta_stops_for_trips(trips=trips.values()))
                # Update trip details
                tasks_trips_details = asyncio.create_task(super()._update_details(trips=trips))

                await task_stops                       
                detailed_trip = await tasks_trips_details

                trains.append(list(detailed_trip.values())[0])

            return trains

        except Exception as e:
            self._logger.error(f"Error updating trips scheduling and info: {e}")
            raise
   
class MBTATripError(Exception):
    pass



from datetime import datetime
import logging
from typing import Any, Optional

class MBTASchedule:
    """A schedule object to hold information about a schedule."""

    def __init__(self, schedule: dict[str, Any]) -> None:
        try:
            # ID
            self.id: Optional[str] = schedule.get('id', None)
            
            # Attributes
            attributes = schedule.get('attributes', {})
            self.timepoint: Optional[bool] = attributes.get('timepoint', None)
            self.stop_sequence: Optional[int] = attributes.get('stop_sequence', None)
            self.stop_headsign: Optional[str] = attributes.get('stop_headsign', None)
            self.pickup_type: Optional[int] = attributes.get('pickup_type', None)
            self.drop_off_type: Optional[int] = attributes.get('drop_off_type', None)
            self.direction_id: Optional[int] = attributes.get('direction_id', None)
            self.departure_time: Optional[datetime] = (
                datetime.fromisoformat(attributes.get('departure_time'))
                if attributes.get('departure_time') is not None
                else None
            )
            self.arrival_time: Optional[datetime] = (
                datetime.fromisoformat(attributes.get('arrival_time'))
                if attributes.get('arrival_time') is not None
                else None
            )

            # Relationships
            relationships = schedule.get('relationships', {})
            self.trip_id: Optional[str] = relationships.get('trip', {}).get('data', {}).get('id', None) if relationships.get('trip', {}).get('data') is not None else None
            self.stop_id: Optional[str] = relationships.get('stop', {}).get('data', {}).get('id', None) if relationships.get('stop', {}).get('data') is not None else None
            self.route_id: Optional[str] = relationships.get('route', {}).get('data', {}).get('id', None) if relationships.get('route', {}).get('data') is not None else None
            self.prediction_id: Optional[str] = relationships.get('prediction', {}).get('data', {}).get('id', None) if relationships.get('prediction', {}).get('data') is not None else None

        
        except Exception as e:
            # Log the exception with traceback
            logger = logging.getLogger(__name__)
            logger.error(f"Error initializing {self.__class__.__name__}: {e}", exc_info=True)

    def __repr__(self) -> str:
        return (f"MBTASchedule:{self.id}")

    def __eq__(self, other: object) -> bool:
        """Defines equality based on the schedule ID."""
        if isinstance(other, MBTASchedule):
            return self.id == other.id
        return False



from datetime import datetime
from enum import Enum
import logging
from typing import Any, Optional

class MBTAPrediction:
    """A prediction object to hold information about a prediction."""
    
    def __init__(self, prediction: dict[str, Any]) -> None:
        try:
            # ID
            self.id: Optional[str] = prediction.get('id', None)
            
            # Attributes
            attributes = prediction.get('attributes', {})
            self.update_type: Optional[str] = attributes.get('update_type', None)
            self.stop_sequence: Optional[int] = attributes.get('stop_sequence', None)
            self.status: Optional[str] = attributes.get('status', None)
            self.schedule_relationship: Optional[str] = attributes.get('schedule_relationship', None)
            self.revenue_status: Optional[str] = attributes.get('revenue_status', None)
            self.direction_id: Optional[int] = attributes.get('direction_id', None)        
            self.departure_uncertainty: Optional[int] = attributes.get('departure_uncertainty', None)
            self.departure_time: Optional[datetime] = (
                datetime.fromisoformat(attributes.get('departure_time'))
                if attributes.get('departure_time') is not None
                else None
            )
            self.arrival_uncertainty: Optional[int] = attributes.get('arrival_uncertainty', None)
            self.arrival_time: Optional[datetime] = (
                datetime.fromisoformat(attributes.get('arrival_time'))
                if attributes.get('arrival_time') is not None
                else None
            )
            
            # Relationships
            relationships: dict = prediction.get('relationships', {})
            self.vehicle_id: Optional[str] = relationships.get('vehicle', {}).get('data', {}).get('id', None) if relationships.get('vehicle', {}).get('data') is not None else None
            self.stop_id: Optional[str] = relationships.get('stop', {}).get('data', {}).get('id', None) if relationships.get('stop', {}).get('data') is not None else None
            self.trip_id: Optional[str] = relationships.get('trip', {}).get('data', {}).get('id', None) if relationships.get('trip', {}).get('data') is not None else None
            self.schedule_id: Optional[str] = relationships.get('schedule', {}).get('data', {}).get('id', None) if relationships.get('schedule', {}).get('data') is not None else None
            self.route_id: Optional[str] = relationships.get('route', {}).get('data', {}).get('id', None) if relationships.get('route', {}).get('data') is not None else None
            # Extract a list of alert IDs
            self.alerts_id: Optional[list[Optional[str]]] = [
                alert.get('id') for alert in relationships.get('alerts', {}).get('data', []) if alert.get('id') is not None
            ]
            
        except Exception as e:
            # Log the exception with traceback
            logger = logging.getLogger(__name__)
            logger.error(f"Error initializing {self.__class__.__name__}: {e}", exc_info=True)

    def __repr__(self) -> str:
        return (f"MBTAPrediction:{self.id}")

    def __eq__(self, other: object) -> bool:
        """Defines equality based on the prediction ID."""
        if isinstance(other, MBTAPrediction):
            return self.id == other.id
        return False

class MBTAScheduleRelationship(Enum):
    ADDED = "ADDED"  # An extra trip added in addition to a running schedule.
    CANCELLED = "CANCELLED"  # A trip that existed in the schedule but was removed.
    NO_DATA = "NO_DATA"  # No data given for this stop; no realtime info available.
    SKIPPED = "SKIPPED"  # A stop that was originally scheduled but was skipped.
    UNSCHEDULED = "UNSCHEDULED"  # A trip running with no schedule associated.



from enum import Enum
import logging
import time
from typing import Optional, Dict, Any, Tuple
import hashlib
import json
from collections import OrderedDict

_LOGGER = logging.getLogger(__name__)

class CacheEvent(Enum):
    HIT = "hit"
    MISS = "miss"
    EVICTION = "eviction"
    UPDATE = "update"

class MBTACacheManager:
    """
    Manages caching with expiration policies for server-side cache.
    """

    DEFAULT_MAX_CACHE_SIZE = 512

    def __init__(
        self,
        max_cache_size: Optional[int] = DEFAULT_MAX_CACHE_SIZE,
        stats: Optional[bool] = True,
        stats_interval: Optional[int] = None,
        logger: Optional[logging.Logger] = None
    ):
        self._max_cache_size = max_cache_size
        self._cache = OrderedDict()  # Use OrderedDict for LRU behavior
        self._logger = logger or logging.getLogger(__name__)
        self.stats = stats
        if stats:
            self.cache_stats = MBTACacheManagerStats(
                max_cache_size=max_cache_size, 
                stats_interval=stats_interval or None,
                logger=logger )
        self._logger.debug("MBTACacheManager initialized")

    @staticmethod
    def generate_cache_key(path: str, params: Optional[Dict[str, Any]]) -> str:
        """Generate a unique cache key based on the path and parameters."""
        key_data = {"path": path, "params": params or {}}
        return hashlib.sha256(json.dumps(key_data, sort_keys=True).encode()).hexdigest()

    def _enforce_cache_size(self) -> None:
        """Ensure the cache does not exceed the maximum size."""
        while len(self._cache) > self._max_cache_size:
            self._cache.popitem(last=False)  # Remove the oldest item (FIFO)
            if self.stats:
                self.cache_stats.increase_counter(CacheEvent.EVICTION)

    def cleanup(self):
        """Clear all cached data."""
        self._logger.debug("Cleaning up MBTACacheManager resources")
        self.cache_stats.print_stats()
        self._cache.clear()

    def get_cached_data(
        self, path: str, 
        params: Optional[Dict[str, Any]]) -> Tuple[Optional[Any],Optional[float],Optional[str]]:
        """Retrieve cached data from the server-side cache."""
        key = self.generate_cache_key(path, params)
        cached_entry = self._cache.get(key)
        if cached_entry:
            if self._is_cache_entry_valid(cached_entry):
                self._cache.move_to_end(key, last=True)  # Move accessed item to the end (LRU)
                return cached_entry["data"], cached_entry["timestamp"], cached_entry["last_modified"]
            else:
                del self._cache[key]
                if self.stats:
                    self.cache_stats.increase_counter(CacheEvent.EVICTION)
        return None, None, None

    def update_cache(
        self,
        path: str,
        params: Optional[Dict[str, Any]],
        data: Any, 
        last_modified: Optional[str] = None) -> float:
        """Update the server-side cache with data."""
        key = self.generate_cache_key(path, params)
        timestamp = time.time()
        self._cache[key] = {
            "data": data,
            "timestamp": timestamp,
            "last_modified": last_modified
        }
        self._enforce_cache_size()
        if self.stats:
            self.cache_stats.increase_counter(CacheEvent.UPDATE,cache_size=len(self._cache))
        return timestamp

    def _is_cache_entry_valid(self, cached_entry: Dict) -> bool:
        """Check if the cache entry is valid."""
        # You can add custom logic here, e.g., check for expiration time
        return True

class MBTACacheManagerStats:

    DEFAULT_STAS_INTERVAL = 1000

    def __init__(
        self,
        max_cache_size: int,
        stats_interval: Optional[int] = DEFAULT_STAS_INTERVAL,
        logger: Optional[logging.Logger] = None,
    ):
        self.stats_interval = stats_interval or self.DEFAULT_STAS_INTERVAL
        self.max_cache_size = max_cache_size
        self._hits = 0
        self._misses = 0
        self._evictions = 0
        self._entries  = 0
        self._logger = logger or logging.getLogger(__name__)

    @property
    def _requests(self) -> Optional[int]:
        return self._hits + self._misses

    def increase_counter(self, cache_event: CacheEvent, cache_size: Optional[int]= None):

        if cache_event == CacheEvent.HIT:
            self._hits += 1
        elif cache_event == CacheEvent.MISS:
            self._misses += 1
        elif cache_event == CacheEvent.UPDATE:
            self._entries = cache_size
        elif cache_event == CacheEvent.EVICTION:
            self._evictions += 1
            self._entries = max(0, self._entries - 1)

        if self.stats_interval != 0 and cache_event in [CacheEvent.HIT,CacheEvent.MISS] and self._requests > 0 and self._requests % self.stats_interval == 0:
            self.print_stats()

    def print_stats(self):

        hit_rate = (
            int((self._hits / self._requests) * 100)
            if self._requests > 0
            else 0
        )
        usage = (
            int((self._entries / self.max_cache_size) * 100)
            if self.max_cache_size > 0
            else 0
        )
        self._logger.info("MBTA Cache Stats:")
        self._logger.info(f"{self._generate_bar(hit_rate)} {hit_rate}% hit rate ({self._hits}/{self._requests})")        
        self._logger.info(f"{self._generate_bar(usage)} {usage}% usage ({self._entries}/{self.max_cache_size})")
        if self._evictions >0:
            self._logger.info(f"{self._evictions} evictions")

    def _generate_bar(self, percentage):
        bar_length = 10
        filled_length = max(0, min(bar_length, int((percentage / 100) * bar_length)))
        bar_content = "█" * filled_length + "░" * (bar_length - filled_length)
        return f"|{bar_content}|"



import logging
from typing import Any, Optional

class MBTARoute:
    """A route object to hold information about a route."""

    ROUTE_TYPES= {
        0: 'Light Rail',   
        1: 'Heavy Rail',  
        2: 'Commuter Rail',
        3: 'Bus',
        4: 'Ferry'
    }
    
    def __init__(self, route: dict[ str, Any]) -> None:
        try:
            # ID
            self.id:  Optional[str] = route.get('id', None)
            
            # Attributes
            attributes = route.get('attributes', {})
            self.type: Optional[str] = attributes.get('type', None)
            self.text_color: Optional[str] = attributes.get('text_color', None)
            self.sort_order: Optional[int] = attributes.get('sort_order', None)
            self.short_name: Optional[str] = attributes.get('short_name', None)
            self.long_name: Optional[str] = attributes.get('long_name', None)
            self.fare_class: Optional[str] = attributes.get('fare_class', None)
            self.direction_names: list[Optional[str]] = attributes.get('direction_names', [])
            self.direction_destinations: list[Optional[str]] = attributes.get('direction_destinations', [])
            self.description: Optional[str] = attributes.get('description', None)
            self.color: Optional[str] = attributes.get('color', None)
       
        except Exception as e:
            # Log the exception with traceback
            logger = logging.getLogger(__name__)
            logger.error(f"Error initializing {self.__class__.__name__}: {e}", exc_info=True)
        
    def __repr__(self) ->  Optional[str]:
        return (f"MBTAroute:{self.id}")

    def __eq__(self, other: object) -> bool:
        """Defines equality based on the route ID."""
        if isinstance(other, MBTARoute):
            return self.id == other.id
        return False
    
    @staticmethod
    def get_route_type_desc_by_type_id(route_type: int) ->  Optional[str]:
        """Get a description of the route type."""
        return MBTARoute.ROUTE_TYPES.get(route_type, 'Unknown')
    




import asyncio
import logging

from datetime import datetime, timedelta
from abc import abstractmethod
import traceback
from typing import Optional, Tuple, Union

from ..mbta_object_store import MBTARouteObjStore, MBTAStopObjStore

from ..client.mbta_client import MBTAClient

from ..trip import Trip
from ..stop import StopType

from ..models.mbta_stop import MBTAStop
from ..models.mbta_schedule import MBTASchedule
from ..models.mbta_prediction import MBTAPrediction, MBTAScheduleRelationship
from ..models.mbta_alert import MBTAAlert, MBTAAlertPassengerActivity, MBTAAlertsInformedEntity

class MBTABaseHandler:

    FILTER_TIME_DEPARTURE_BUFFER = timedelta(seconds=90)
    FILTER_TIME_ARRIVAL_BUFFER = timedelta(seconds=60)
    DEFAULT_MAX_TRIPS = 1

    def __init__(
        self,
        mbta_client: MBTAClient,
        max_trips: Optional[int],
        logger: Optional[logging.Logger]):

        self._mbta_client = mbta_client
        self._mbta_stops_ids: dict[StopType, list[str]] = {
            StopType.DEPARTURE: [],
            StopType.ARRIVAL: []
        }
        self._max_trips = max_trips
        self._mbta_trip_stops_ids = set()

        #self._trips: dict[str, Trip] = {}  # Dictionary to store Trip objects, keyed by trip ID

        self._last_processed_scheduling = {
             'data': None,
             'timestamp': None
        }

        self._logger = logger or logging.getLogger(__name__)  # Logger instance

    @classmethod
    async def _create(
        cls,
        mbta_client: MBTAClient,
        departure_stop_name: Optional[str] = None ,
        arrival_stop_name: Optional[str] = None,
        max_trips: Optional[int] = DEFAULT_MAX_TRIPS,
        logger: Optional[logging.Logger] = None)-> "MBTABaseHandler":

        instance: MBTABaseHandler = cls(mbta_client=mbta_client, max_trips=max_trips,logger=logger)
        await instance.__update_mbta_stops(departure_stop_name=departure_stop_name,arrival_stop_name=arrival_stop_name)
        return instance

    @abstractmethod
    async def update(self) -> list[Trip]:
        """Update the trips info."""

    ### MBTA STOP
    async def __update_mbta_stops(
        self,
        departure_stop_name: str,
        arrival_stop_name: Optional[str]) -> None:

        self._logger.debug("Updating MBTA stops")

        try:
            mbta_stops, _ = await self.__fetch_mbta_stops()
            self.__process_mbta_stops(mbta_stops, departure_stop_name,arrival_stop_name)

        except Exception as e:
            self._logger.error(f"Error updating MBTA stops: {e}")
            raise

    async def __fetch_mbta_stops(self, params: dict = None) -> Tuple[list[MBTAStop],float]:
        base_params = {
            'filter[location_type]': '0'
            }

        if params is not None:
            base_params.update(params)

        mbta_stops, timestamp = await self._mbta_client.fetch_stops(base_params)
        return mbta_stops, timestamp


    def __process_mbta_stops(
        self,
        mbta_stops: list[MBTAStop],
        departure_stop_name: Optional[str],
        arrival_stop_name: Optional[str]) -> None:

        for mbta_stop in mbta_stops:
            if departure_stop_name and departure_stop_name.lower() == mbta_stop.name.lower() :
                MBTAStopObjStore.store(mbta_stop)
                self._mbta_stops_ids[StopType.DEPARTURE].append(mbta_stop.id)
            if arrival_stop_name and arrival_stop_name.lower() == mbta_stop.name.lower():
                MBTAStopObjStore.store(mbta_stop)
                self._mbta_stops_ids[StopType.ARRIVAL].append(mbta_stop.id)

        if departure_stop_name and len(self._mbta_stops_ids[StopType.DEPARTURE]) == 0:
            self._logger.error(f"Invalid departure stop name, no MBTA stop name matching {departure_stop_name} ")
            raise MBTAStopError(f"Invalid departure stop name, no MBTA stop name matching {departure_stop_name}")

        if arrival_stop_name and len(self._mbta_stops_ids[StopType.ARRIVAL]) == 0:
            self._logger.error(f"Invalid arrival stop name, no MBTA stop name matching {arrival_stop_name} ")
            raise MBTAStopError(f"Invalid arrival stop name, no MBTA stop name matching {arrival_stop_name}")

    ### SCHEDULING
    async def _update_scheduling(
        self,
        trips: dict[str, Trip],
        params: Optional[dict] = None) -> dict[str, Trip] :

        self._logger.debug("Updating scheduling")

        try:

            ####
            params_predictions = None
            if params and "filter[date]" in params.keys():
                params_predictions = params.copy()
                del params_predictions["filter[date]"]
            ####

            task_schedules = asyncio.create_task(self.__fetch_schedules(params))
            task_predictions = asyncio.create_task(self.__fetch_predictions(params_predictions))

            mbta_schedules, timestamp = await task_schedules

            if self._last_processed_scheduling['timestamp'] != timestamp:
                self._logger.debug("New MBTA schedules data detected. Processing...")
                trips = await self.__process_scheduling(schedulings=mbta_schedules,trips=trips)
                self._last_processed_scheduling['data'] = trips
                self._last_processed_scheduling['timestamp'] = timestamp
            else:
                self._logger.debug("MBTA Schedules data are up-to-date. Skipping processing.")
                trips = self._last_processed_scheduling['data']

            mbta_predictions, _ = await task_predictions

            trips = await self.__process_scheduling(schedulings=mbta_predictions, trips=trips)

            return trips

        except Exception as e:
            self._logger.error(f"Error updating scheduling: {e}")
            raise

    async def __fetch_schedules(
        self,
        params: Optional[dict] = None) -> Tuple[list[MBTASchedule],float]:

        base_params = {
            'filter[stop]': ','.join(
                self._mbta_stops_ids.get(StopType.DEPARTURE, []) +
                self._mbta_stops_ids.get(StopType.ARRIVAL, [])
            ),
            'sort': 'departure_time',
        }

        if params is not None:
            base_params.update(params)

        mbta_schedules, timestamp = await self._mbta_client.fetch_schedules(base_params)
        return mbta_schedules, timestamp

    async def __fetch_predictions(
        self,
        params: Optional[dict] = None) -> Tuple[list[MBTAPrediction],float]:

        base_params = {
            'filter[stop]': ','.join(
                self._mbta_stops_ids.get(StopType.DEPARTURE, []) +
                self._mbta_stops_ids.get(StopType.ARRIVAL, [])
            ),
            'filter[revenue]': 'REVENUE',
            'sort': 'departure_time'
        }

        if params is not None:
            base_params.update(params)

        mbta_predictions, timestamp = await self._mbta_client.fetch_predictions(base_params)
        return mbta_predictions, timestamp

    async def __process_scheduling(self, schedulings: Union[list[MBTASchedule],list[MBTAPrediction]], trips: dict[str, Trip] = None ) -> dict[str, Trip] :
        self._logger.debug("Processing scheduling")

        try:

            for scheduling in schedulings:

                # if schedulings is a prediciton and ScheduleRelationship == specific cases
                if isinstance(scheduling, MBTAPrediction) and scheduling.schedule_relationship in [MBTAScheduleRelationship.CANCELLED.value, MBTAScheduleRelationship.SKIPPED.value, MBTAScheduleRelationship.NO_DATA.value]:
                    # if the scheduling stop is already in the trips
                    if  scheduling.trip_id in trips:
                        # remove it
                        trips.pop(scheduling.trip_id)
                    # skip this scheduling
                    continue
                
                # If the trip of the prediction is not in the trips dict
                if scheduling.trip_id not in trips:
                    # Create the trip
                    trip = Trip()
                else:
                    # Or get the trip with same trop_id from the trips
                    trip = trips[scheduling.trip_id]
                    
                #Set the mbta_route for the trip, 
                if not trip.mbta_route and scheduling.route_id:
                    # if the route in the objStore
                    if MBTARouteObjStore.get_by_id(scheduling.route_id):
                        # set the route id for the trip
                        trip._mbta_route_id = scheduling.route_id
                    else:
                        # fetch the route
                        mbta_route, _ = await self._mbta_client.fetch_route(scheduling.route_id)
                        # set the route for the trip (it will set the route id and store the obj in the objstore)
                        trip.mbta_route = mbta_route
                
                # if route type 0 or 1 and MBTASchedule skipp
                #https://www.mbta.com/developers/v3-api/best-practices#predictions
                if trip.mbta_route.type in [0,1] and isinstance(scheduling, MBTASchedule):
                    #Unfortunately in some corner cases there may be more trips type for the same trip...cannot break
                    continue
                    
                # Check if the prediction stop_id is in the departure stops lists
                if scheduling.stop_id in self._mbta_stops_ids[StopType.DEPARTURE]:
                    #add/update the stop as departure
                    trip.add_stop(
                        stop_type=StopType.DEPARTURE,
                        scheduling=scheduling,
                        mbta_stop_id=scheduling.stop_id)
                #Check if the prediction stop_id is in the arrival stops lists
                elif scheduling.stop_id in self._mbta_stops_ids[StopType.ARRIVAL]:
                    #add/update the stop as arrival
                    trip.add_stop(
                        stop_type=StopType.ARRIVAL,
                        scheduling=scheduling,
                        mbta_stop_id=scheduling.stop_id)
                    
                trips[scheduling.trip_id] = trip
                    
            return trips
        
        except Exception as e:
            self._logger.error(f"Error while processing scheduling: {e}")
            self._logger.error(traceback.format_exc())  # Log the stack trace
            raise  # Re-raise the exception

    def _get_mbta_stop_by_id(self, stop_id: str) -> Optional[MBTAStop]:
        if any(stop_id in stop_ids for stop_ids in self._mbta_stops_ids.values()):
            return MBTAStopObjStore.get_by_id(stop_id)
        return None
    
        
    ## SET TRIP DETAILS
    async def _update_details(self, trips: dict[str, Trip]) -> dict[str, Trip]:
        """Update trip details with MBTA trip, vehicle, and alert information."""
        self._logger.debug("Updating trips details")
        
        try:
            updated_trips: dict[str, Trip] = {}

            # Collect all trip IDs to fetch vehicles
            params = {'filter[trip]': ','.join(trips.keys())}
            mbta_stop_ids = self._mbta_stops_ids.get(StopType.DEPARTURE, []) + self._mbta_stops_ids.get(StopType.ARRIVAL, [])
            
            # Fetch vehicles and alerts concurrently
            task_vehicles = asyncio.create_task(self._mbta_client.fetch_vehicles(params))
            task_alerts = asyncio.create_task(self.__get_mbta_alerts(trips=trips, stops_ids=mbta_stop_ids))

            # Await results from both tasks
            mbta_vehicles, _ = await task_vehicles
            mbta_alerts, _ = await task_alerts

            for trip_id, trip in trips.items():
                # # Assign the trip ID if not already set
                # if not trip.mbta_trip and not trip.mbta_trip.id:
                #     trip.mbta_trip.id = trip_id

                # Fetch and assign the MBTA trip if not already set
                if not trip.mbta_trip:
                    mbta_trip, _ = await self._mbta_client.fetch_trip(id=trip_id)
                    trip.mbta_trip = mbta_trip

                # Match vehicle for the trip
                matching_vehicles = [vehicle for vehicle in mbta_vehicles if vehicle.trip_id == trip_id]

                # Assign the vehicle to the trip if exist
                trip.mbta_vehicle = matching_vehicles[0] if len(matching_vehicles) > 0 else None

                # Filter and sort relevant alerts
                processed_alerts = []
                for mbta_alert in mbta_alerts:
                    if mbta_alert not in processed_alerts and mbta_alert.severity > 1 and self.__is_alert_relevant(mbta_alert, trip):
                        processed_alerts.append(mbta_alert)
                        
                # Sort processed_alerts by severity (high to low)
                processed_alerts.sort(key=lambda mbta_alert: mbta_alert.severity, reverse=True)

                trip.mbta_alerts = processed_alerts

                # Add the updated trip to the results
                updated_trips[trip_id] = trip

            return updated_trips

        except Exception as e:
            self._logger.error(f"Error updating trips details: {e}")
            raise


    async def __get_mbta_alerts(
        self,
        trips: Optional[dict[str, Trip]] = None, 
        stops_ids: Optional[list[str]] =  None)-> Tuple[list[MBTAAlert],str]:
        self._logger.debug("Updating MBTA alerts")
        try:

            trips_ids = list(trips.keys())

            params = {
                'filter[stop]': ','.join( stops_ids ),
                'filter[trip]': ','.join( trips_ids ),
            }

            mbta_alerts, timestamp = await self._mbta_client.fetch_alerts(params)

            return mbta_alerts, timestamp

        except Exception as e:
            self._logger.error(f"Error updating MBTA alerts: {e}")
            raise
        
    def __is_alert_relevant(self, mbta_alert: MBTAAlert, trip: Trip) -> bool:
        """
        Determines whether an MBTA alert is relevant to a specific trip.

        Args:
            mbta_alert (MBTAAlert): The alert object containing details such as informed entities and active period.
            trip (Trip): The trip object containing information about the route, stops, and timing.

        Returns:
            bool: True if the alert is relevant to the given trip, otherwise False.

        Relevance is determined based on the following criteria:
            - The alert matches the trip's route (route-level alert).
            - The alert's informed entity explicitly mentions the trip ID.
            - The alert's informed entity is tied to the departure or arrival stops with relevant activities (boarding or exiting).
            - The trip's departure or arrival time falls within the alert's active period.
        """
        
        try:
            
            def check_route_level_alert(informed_entity: MBTAAlertsInformedEntity):
                """
                Check if the informed entity is a route-level alert matching the trip's route and direction.

                Args:
                    informed_entity (MBTAAlertsInformedEntity): The informed entity to check.

                Returns:
                    bool: True if the entity matches the route-level alert criteria, otherwise False.
                """
                return (
                    informed_entity.route_id == trip.mbta_route.id and
                    not informed_entity.stop_id and
                    not informed_entity.trip_id and
                    (not informed_entity.direction_id or informed_entity.direction_id == trip.mbta_trip.direction_id)
                )
            
            trip_id = trip.mbta_trip.id if trip.mbta_trip else None
            departure_stop_id = trip.get_stop_id_by_stop_type(StopType.DEPARTURE)
            arrival_stop_id = trip.get_stop_id_by_stop_type(StopType.ARRIVAL)

            for informed_entity in mbta_alert.informed_entities:
                
                active_period_start = mbta_alert.active_period_start.replace(tzinfo=None)
                active_period_end = mbta_alert.active_period_end.replace(tzinfo=None)
                
                try:
                    if (
                        check_route_level_alert(informed_entity) or
                        informed_entity.trip_id == trip_id or
                        (
                            informed_entity.stop_id == departure_stop_id and 
                            MBTAAlertPassengerActivity.BOARD in informed_entity.activities
                        ) or
                        (
                            informed_entity.stop_id == arrival_stop_id and 
                            MBTAAlertPassengerActivity.EXIT in informed_entity.activities
                        )
                    ) and (
                        (
                            trip.departure_time and 
                            (active_period_start is None or active_period_start <= trip.departure_time) and 
                            (active_period_end is None or trip.departure_time <= active_period_end)
                        ) or (
                            trip.arrival_time and 
                            (active_period_start is None or active_period_start <= trip.arrival_time) and 
                            (active_period_end is None or trip.arrival_time <= active_period_end)
                        )
                    ):
                        return True

                except Exception as e:
                    logging.error(f"Exception processing informed_entity {informed_entity}: {e}")
                    logging.error(traceback.format_exc())  # Logs full stack trace
                    continue  # Skip this entity and continue processing others

        except Exception as e:
            logging.error(f"Unexpected error in __is_alert_relevant: {e}")
            logging.error(traceback.format_exc())  # Logs full stack trace

        return False

    ##UTILY METHODS FOR SUBCLASSES
    def _filter_and_sort_trips(
        self,
        trips: dict[str, Trip],
        remove_departed: Optional[bool] = True,
        require_both_stops: Optional[bool] = True,
        sort_by: Optional[StopType] = StopType.DEPARTURE) -> dict[str, Trip]:
       
        """Filter and sort trips based on conditions like direction, departure, and arrival times."""
        self._logger.debug("Filtering Trips")
        now = datetime.now().astimezone()
        filtered_trips: dict[str, Trip] = {}

        try:
   
            trips = self._sort_trips(trips, sort_by)
            
            for trip_id, trip in trips.items():
                departure_stop = trip.get_stop_by_type(StopType.DEPARTURE)
                arrival_stop = trip.get_stop_by_type(StopType.ARRIVAL)

                # Ensure both departure and arrival stops exist and are in the correct sequence
                if require_both_stops and (
                    not departure_stop or not arrival_stop or departure_stop.stop_sequence > arrival_stop.stop_sequence
                ):
                    continue
                else:
                    # if sequence cannot be used, check that a DEPARTURE stop has departure Time obj (is not the last stop of a trip) and ARRIVAL stop has arrival Time obj (is not the first stop of a tirp)
                    if departure_stop and not departure_stop.departure:
                        continue
                    if arrival_stop and not arrival_stop.arrival:
                        continue

                # Remove trips that have already departed (+BUFFER time)
                if remove_departed and departure_stop.departure_time and departure_stop.departure_time < now - self.FILTER_TIME_DEPARTURE_BUFFER:
                    continue

                # Remove trips that have already arrived (+BUFFER time)
                if arrival_stop and arrival_stop.arrival_time and arrival_stop.arrival_time < now - self.FILTER_TIME_ARRIVAL_BUFFER:
                    continue

                # If vehicle data is available, use it for filtering
                if trip.mbta_vehicle:

                    if remove_departed and departure_stop and trip.has_departed:
                        continue  # If the vehicle has passed the departure stop, filter it out

                    if arrival_stop and trip.has_arrived:
                        continue  # If the vehicle has passed the arrival stop, filter it out

                # Add the valid trip to the processed trips
                filtered_trips[trip_id] = trip

            return dict(list(filtered_trips.items())[:(self._max_trips * 2)])

        except Exception as e:
            self._logger.error(f"Error filtering trips: {e}")
            raise

    def _sort_trips(
        self,
        trips: dict[str, Trip],
        sort_by: Optional[StopType]=StopType.DEPARTURE) -> dict[str, Trip]:

        self._logger.debug("Cleaning Trips")
        try:

            today = datetime.now()
            sorted_trips: dict[str, Trip] = {
                trip_id: trip
                for trip_id, trip in sorted(
                    trips.items(),
                    key=lambda item: (
                        item[1].get_stop_by_type(sort_by).time 
                        if item[1].get_stop_by_type(sort_by) and item[1].get_stop_by_type(sort_by).time is not None
                        else (today + timedelta(days=365)).astimezone()
                    )
                )
            }

            return sorted_trips

        except Exception as e:
            self._logger.error(f"Error sorting and cleaning trips: {e}")
            raise

    async def _update_mbta_stops_for_trips(self, trips: list[Trip]) -> None:
        
        for trip in trips:
            params = {
                    'filter[route]': trip.mbta_route.id,
                    'include' : 'child_stops'
            }
            mbta_stops, _ = await self._mbta_client.fetch_stops(params=params)
            for mbta_stop in mbta_stops:
                if mbta_stop.id not in self._mbta_stops_ids or not MBTAStopObjStore.get_by_id(mbta_stop.id):
                    self._mbta_trip_stops_ids.add(mbta_stop.id)
                    MBTAStopObjStore.store(mbta_stop)
                    
class MBTAStopError(Exception):
    pass



import logging
from typing import Any, Optional

class MBTATrip:
    """A trip object to hold information about a trip."""
    
    def __init__(self, trip: dict[str, Any]) -> None:
        try:
            # ID
            self.id: str = trip.get('id', '')
            
            # Attributes
            attributes = trip.get('attributes', {})
            self.wheelchair_accessible: Optional[bool] = attributes.get('wheelchair_accessible', None)
            self.revenue_status: Optional[bool] = attributes.get('revenue_status', None)
            self.name: Optional[str] = attributes.get('name', None)
            self.headsign: Optional[str] = attributes.get('headsign', None)
            self.direction_id: Optional[int] = attributes.get('direction_id', None)
            self.block_id: Optional[str] = attributes.get('block_id', None)
            self.bikes_allowed: Optional[bool] = attributes.get('bikes_allowed', None)
            
            # Relationships
            relationships = trip.get('relationships', {})
            self.route_id: Optional[str] = relationships.get('route', {}).get('data', {}).get('id', None)  if relationships.get('route', {}).get('data') is not None else None

        except Exception as e:
            # Log the exception with traceback
            logger = logging.getLogger(__name__)
            logger.error(f"Error initializing {self.__class__.__name__}: {e}", exc_info=True)

    def __repr__(self) -> str:
        return (f"MBTATrip:{self.id}")

    def __eq__(self, other: object) -> bool:
        """Defines equality based on the trip ID."""
        if isinstance(other, MBTATrip):
            return self.id == other.id
        return False
